//breastSculptV2 by stuntcock

import flash.net.URLRequest;
import flash.geom.Point;
import flash.geom.Transform;
import com.ryan.geom.FreeTransformEvent;
import flash.geom.Rectangle;
import flash.display.Sprite;
import flash.display.DisplayObject;
import flash.geom.Matrix;
import flash.display.DisplayObjectContainer;
import flash.utils.Dictionary;

// The following are basic variables which will store fundamental pointers.
// They provide us with a "gateway" into SDT data.
var main;
var modSettingsLoader:Class;
var g;
var her;

// Note that we've deliberately included the version number in the settings filename.
// This should reduce the risk that users will inadvertently continue to use an old 
// settings file after upgrading the mod (and then complain that new features are 
// missing or disabled!)
var setsettingsfilename:String;
var modName:String;

// Config setting.  Defines which hotkey will initiate (and terminate) sculpting mode.
var keytoactivate:Number;
var keytoactivateSkew:Number;
var keytocancel:Number;
var keytorecenter:Number;
// Config setting.  If 1, breast shape will be set to the largest size available (within the
// scope of the default SDT slider, of course) during initialization.
var autosettomaxbreastsize:Number;
// Config setting.  If 1, then a sculpting operation can be initiated by clicking the breasts.
var allowMouseActivation:Number;
// Config setting.  If nonnegative, it will be applied during initialization.
var breastFirmnessOverride:Number;
// Config setting.  If 1, minor visual elements of the scene will be concealed.
var hideInframammaryFold:Number;
// Config setting.  If 1, custom shapes will be preserved when no replacement exists.
var retainCustomSculpting:Number;
// Config setting.  If 1, saved character data will include sculpting parameters.
var includeSculptDataInCharacter:Number;
// Config setting.  If 1, all transformation handles will be shown.
var expertMode:Number;
// Config settings, which control the appearance of the transform UI controls.
var transformLineColor:uint;
var transformFillColor:uint;
var transformLineThickness:Number;
var transformHandleSize:Number;
var transformBoundingBoxColor:uint;
// Config setting.  If 1, the mod is enabled.  If any other value, the mod will stop working
// (or abort itself during initialization).
var breastSculpt;

// Reference to the third-party transformation module which will be invoked by the mod.
// Expansion note: Only one transformation manager is needed, even if we eventually allow 
// users to transform multiple objects within the scene (breasts, buttocks, noses, etc).
var anatomyTransform:com.ryan.geom.FreeTransformManager;
// The following variables store "anatomy" references (pointers to DisplayObjects).
// They will be populated during initialization, and then used in "background" code
// to simplify the repetitive "maintenance" activities which occur constantly during
// normal gameplay.
var leftBreast:DisplayObject;
var rightBreast:DisplayObject;
// For each "sculptable" object in the scene, we must store a matrix which contains
// the current "desired changes."  The matrix can be applied to the scene whenever 
// the sculpting effect must be applied or refreshed.
var mostRecentTransform_LeftBreast:Matrix;
var mostRecentTransform_RightBreast:Matrix;
// The breast sprite does not rest at the (0,0) origin position of its parent (torso)
// but is instead offset somewhat.  We must know this offset so that we can correctly
// position the bounding box.  Unfortunately, we cannot just look it up at runtime
// (because we will actively CHANGE its offset) so we must detect and store this value
// during mod initialization.
var breastOffsetMatrix:Matrix;

// This is a special proxy.  Most proxies are declared inline and then allowed to
// fall out-of-scope.  This one is declared as an instance because we need to
// repeatedly attach and detach it during the mod's operation.
// It governs the "SaveData" function of the InGameMenu class.
var getSaveDataString;

// The following constants define settings headers/markers which are used when
// appending mod-specific entries to Character Data (and when reading it back).
const charData_LB = "BreastSculpt_LeftBreast";
const charData_RB = "BreastSculpt_RightBreast";
const charData_IM = "BreastSculpt_HideInframammary";

function initl(l)			//boiler plate setup for getting settings started, put startup stuff here that are not dependent on loaded settings
{
	// Initialize the standard "reference" variables
	main = l;
	g = l.g;
	her = l.her;
	
	// Create references to the most frequently-used parts of HER anatomy (for convenience)
	leftBreast = her.torsoBack.leftBreast;
	rightBreast = her.torso.midLayer.rightBreast;
	
	// Apply default values to the configuration parameters
	setsettingsfilename = "breastsculptV2settings";
	modName = "breastsculptV2";
	keytoactivate = 190;			// 190 = "."
	keytorecenter = 188;			// 188 = ","
	keytoactivateSkew = -1;		// -1 = feature disabled
	keytocancel = 27;			//  27 = ESC
	autosettomaxbreastsize = -1;
	allowMouseActivation = -1;
	breastFirmnessOverride = -1;
	hideInframammaryFold = -1;
	retainCustomSculpting = 0;
	includeSculptDataInCharacter = 1;
	expertMode = 1;
	transformLineColor = 0x0000FF;
	transformFillColor = 0xFFFFFF;
	transformLineThickness = 1;
	transformHandleSize = 10;
	transformBoundingBoxColor = 0x5555FF;
	breastSculpt = 1;
	
	// We must capture the breast offset coordinates (relative to torso) before any significant
	// changes are made.  This will allow us to properly cancel out the offset when performing
	// complex transformations.
	breastOffsetMatrix = new Matrix(1, 0, 0, 1, rightBreast.x, rightBreast.y);

	// Prepare to load custom settings from the txt file
	var msl = new modSettingsLoader(setsettingsfilename,settingsLoaded);
	msl.addEventListener("settingsNotFound",settingsNotFound);
	
	// TODO: test out sby's suggestion for inter-mod datalink.
}

// This function is essentially a "callback" - it is invoked when the loader framework
// code has finished reading the settings file that was "requested" by this mod.
// We must use the information provided in the callback in order to populate local variables,
// because it will fall out of scope and we don't want to deal with the hassle of
// re-opening the config file during runtime.
function settingsLoaded(e)					//loads settings from file
{
	// The contents of the settings file will have been parsed into key-value pairs,
	// which is very convenient for us.
	var dict:Dictionary = e.settings;
	// We will iterate through the dictionary.  We'll assume that the key is the name
	// of a local variable, and attempt to assign the value to that variable.
	for (var key:Object in dict)
	{
		var val:Object = dict[key];
		if ( this.hasOwnProperty( key ) )			//checks if variable exists
		{
			this[key] = val;    								//replaces all switch statement variable assignments
		}
		else
		{
			main.updateStatusCol("invalid breastSculpt setting: "+key,"#FF0000");	//displays if there is a setting in the settings file that does not match any variables
			// This error statement helps to warn users that their settings file
			// contains an invalid key.  It may be that they accidentally renamed
			// a setting while altering a value, but perhaps there's a more serious
			// problem (e.g. using a settings file from an obsolete version of the mod).
		}
		// Note: we do NOT warn users about entries that are *missing* from their
		// settings file.  In such cases, the setting will simply retain its default
		// value (such as -1) and its default behaviour (such as "feature disabled").
	}
	
	finishinit();
	main.updateStatusCol(setsettingsfilename+" done loading","#00ff00");
}

function settingsNotFound(e)
{
	main.updateStatusCol(e.msg,"#FF0000");		//display error message
	finishinit();
}

function finishinit()		//put startup stuff here that require loaded settings
{
	if(breastSculpt == 1)
	{
		// Register the hotkeys which will be used to call up the resize/rescale UI elements
		if (keytoactivate > -1) {
			main.registerFunction(toggleBreastSculpting_Proportionate, keytoactivate);
		}
		if (keytoactivateSkew > -1) {
			main.registerFunction(toggleBreastSculpting_Disproportionate, keytoactivateSkew);
		}
		if (keytocancel > -1) {
			main.registerFunction(cancelBreastSculpting, keytocancel);
		}
		if (keytorecenter > -1 && expertMode) {
			// Note: the recentering hotkey is meaningful only if the user is operating
			// in expert mode (in which all of the UI handles, including the center
			// point, are visible).  Otherwise we just ignore it to avoid confusion.
			main.registerFunction(toggleBreastSculpting_Recenter, keytorecenter);
		}
		
		// Proxy the breastController.update function so that we can take reactive/
		// corrective action whenever it gets invoked.
		var lProxy:Class = main.lDOM.getDefinition("Modules.lProxy") as Class;
		var leftBreastUpdate = (lProxy as Class).createProxy(her.leftBreastController, "update");
		var rightBreastUpdate = (lProxy as Class).createProxy(her.rightBreastController, "update");
		leftBreastUpdate.addPost(leftBreastUpdate_Post, true);
		rightBreastUpdate.addPost(rightBreastUpdate_Post, true);
		
		// Proxy the her.breastSizeChanged function so that we can control
		// visibility of the inframammary fold.
		g.characterControl.addBreastSizeChangeListener(breastSizeChanged_Post);
		//var breastSizeUpdate = (lProxy as Class).createProxy(her, "breastSizeChanged");
		//breastSizeUpdate.addPost(breastSizeChanged_Post, true);
		// TODO: the standard lProxy approach is ineffective (don't know why)
		//       so we're using a simpler approach for now.  Investigate!

		// Proxy the Generate Character Data function so that we can insert mod params
		try {
			var getSwfModsDataString = (lProxy as Class).createProxy(g.customElementLoader, "getSwfModsDataString");
			getSwfModsDataString.addPre(getSwfModsDataString_Pre, true);
			// Note that the next line uses an instance variable rather than an inline
			// variable.  That's because we need to detach and reattach the proxy during
			// gameplay (rather than leaving it always-on).
			getSaveDataString = (lProxy as Class).createProxy(g.inGameMenu, "getSaveDataString");
			getSaveDataString.addPre(getSaveDataString_Pre, true);
			var loadData = (lProxy as Class).createProxy(g.inGameMenu, "loadData");
			loadData.addPre(loadData_Pre, true);
			// TODO: adjust/fix/augment this proxy hook so that we can catch Vanilla-character
			//       load events also.
		} catch (myError:Error) {
			main.updateStatusCol("ERR: " + myError.toString(),"#FF0000"); 
		}
		
		try {
			//var resetCharClicked = (lProxy as Class).createProxy(main, "resetCharClicked");
			//resetCharClicked.addPost(resetCharClicked_Post, true);
			// TODO: this doesn't work; maybe we're not allowed to proxy the Loader's own functions?
		} catch (myError:Error) {
			main.updateStatusCol("ERR: " + myError.toString(),"#FF0000"); 
		}

		// Initialize the sculpt matrices (default/identity)
		mostRecentTransform_LeftBreast = new Matrix();
		mostRecentTransform_RightBreast = new Matrix();

		// Apply the custom breast firmness value (if one was provided)
		if(breastFirmnessOverride >= 0) 
		{
			her.leftBreastController._breastFirmness = her.rightBreastController._breastFirmness = breastFirmnessOverride;
		}
		
		// This is a legacy feature from sby's superbreastV2 mod.
		// Because superbreast was intended mainly for size *increases*, it looked a bit
		// silly when used with the smaller breast shapes (e.g. A-cup).  You'd get the desired
		// amount of prominence, but the effect wasn't very convincing.
		// breastsculpt can actually achieve decent results when upsizing the smaller shapes,
		// so long as the player is willing to make a few more adjustments (rotation, skew, 
		// and repositioning).  Nonetheless, if you're trying to achieve giant boobs
		// then it makes sense to start with the largest and roundest possible option.
		if(autosettomaxbreastsize == 1)
		{
			// Max out the breast setting
			g.characterControl.setBreasts(149);
			// Update the slider to reflect this change
			g.inGameMenu.updateBreastSlider();
			// Update any spit/semen strands that fall within the (enlarged) breast area
			g.strandControl.checkElementAnchors(rightBreast);
		}
		
		
		// The following code uses the free-transform-manager-as3 package, found at:
		// https://github.com/ryantan/free-transform-manager-as3
		
		// Create the transformation controller.  It is a persistent object (that is -
		// we create it once per game session, rather than once per transform action).
		anatomyTransform = new com.ryan.geom.FreeTransformManager();
		
		// Customize the basic transformation settings, for the sake of visibility.
		anatomyTransform.boundingBoxOutlineThickness = transformLineThickness;
		anatomyTransform.handleRadius = transformHandleSize;
		anatomyTransform.handleOutlineThickness = transformLineThickness;
		anatomyTransform.handleFillColor = transformFillColor;
		anatomyTransform.handleOutlineColor = transformLineColor;
		anatomyTransform.boundingBoxOutlineColor = transformBoundingBoxColor;
		anatomyTransform.UseSimpleHandles = (expertMode == 0);

		// By default, the transformation controller will adjust the z-order of the 
		// scene; any object which acquires focus will be brought to the front.
		// To avoid unintended disruption of SDT gameplay (and/or mod compatibility
		// problems), we will disable this feature.
		anatomyTransform.bringTargetToTop = false;
		// As a consequence, the bounding box (and its drag handles) can be occulted
		// by other sprites (such as the male model).  This is considered an acceptable
		// risk/deficiency.
		
		// Begin listening for transformation events.
		anatomyTransform.addEventListener(FreeTransformEvent.ON_TRANSFORM, onTransform);
		anatomyTransform.addEventListener(FreeTransformEvent.ON_CANCEL, onTransform);
		anatomyTransform.addEventListener(FreeTransformEvent.ON_FOCUS, onFocus);
		
		// Note that this "listening" step (and our reaction to each event) is crucial,
		// for two reasons:
		// 1) When a transformation action occurs, it will not actually modify the SDT 
		//    scene contents.  Instead, it will generate an event which informs listeners
		//    (e.g. THIS CLASS) of the details.  We must actually decide whether the transform
		//    is valid/appropriate, and then apply it to the scene (assuming that it is).
		// 2) After we make a change to the scene, the SDT core logic will very quickly 
		//    "overwrite" it.  We need to store the transformation data so that it can
		//    remain a persistent aspect of what's happening on-stage.  This persistence
		//    is implemented by the lProxy class - each time that SDT core logic "fixes"
		//    the transformation matrix, we react and "unfix" it.
		
		
		// The torso's middle layer serves as the attachment point for two children:
		// 0 - chest (container)
		// 1 - rightBreast (sprite)
		// The chest serves as the attachment point for:
		// 0 - skin (sprite)
		// 1 - inframammary fold line (sprite)
		// 2 - inframammary fold tan (sprite)
		
		// The inframammary fold is a visual element; it's essentially a dark "shading"
		// and border segment which is drawn on the torso, below the lower curve of the 
		// foreground breast (rightBreast).  It helps to create the visual appearance of
		// a heavy/hanging breast, adding a bit of emphasis on the otherwise-smooth chest.
		
		// However, this mod makes it possible to perform manipulations which will separate
		// the breast from the inframammary fold, leaving the fold itself as just an ugly
		// "scar" or discoloration across the chest.  This effect can be avoided by
		// sticking with conventional shapes, but we prefer to give users greater freedom.
		
		// Once the user has achieved a particular breast shape which suits a character,
		// it may be very difficult to get it to align with BOTH the collarbone and the
		// inframammary fold.
			
		// Therefore, we give users the option to simply hide the inframammary fold.
		helper_InframammaryVisibility();
		
		// The inframmamary fold exists on the right side (foreground) only; we do
		// not need to make a similar adjustment on the left (background) side.


		// HOOK: auto-loading goes here.
		// COMMENT: the "auto-loading" feature will probably not be implemented.
		// The idea was that the settings.txt file could include a comma-separated
		// matrix definition, which would be applied to the female character during 
		// mod loading.  Since we now have the ability to auto-apply transformation
		// during CHARACTER loading (with each transformation being custom-tailored
		// to the character in question), there's no real need for a one-size-fits-all
		// transformation.  Especially one which would need to be defined and tweaked
		// via a set of comma-separated numbers in a settings file (not user-friendly!)


		// At this point, it is possible to "register" a sprite for transformation eligibility
		// By doing so, we would allow users to initiate a transformation simply by clicking
		// (or click-and-drag) on a target sprite (e.g. breast) rather than by pressing the hotkey.
		
		// This feature is disabled by default, because it is expected that it could activate 
		// suddenly and unexpectedly (e.g. while Ctrl-clicking to remove clothing, or during
		// bukkake mode) and thus interfere with the user experience.
		
		// If the registration does *not* occur, then we must initiate "transformation mode"
		// manually (in code) or transformations will never occur.  This is done by listening
		// for hotkeys.  Refer to the top of this function for details.
		if (allowMouseActivation == 1) anatomyTransform.registerSprite(rightBreast);
		anatomyTransform.mouseEnabled = (allowMouseActivation == 1);
	}
	
	// vvv The stuff below is boilerplate for SDT mods vvv
	main.unloadMod();
	main.registerUnloadFunction(doUnload);
	// NOTE: this mod NO LONGER includes a "addEnterFramePersist" call, because we're instead
	// using lProxy to react to specific function calls (instead of blindly executing code
	// off during every single frame of gameplay).
}

// This function is invoked when the user attempts to initiate a sculpting operation
// by clicking directly on a DisplayObject (such as a breast).
function onFocus(evt:FreeTransformEvent):void
{
	// Initiate a sculpting operation, just as if the user had pressed the hotkey.
	toggleBreastSculpting_Proportionate();
}

// This function is invoked whenever the FreeTransformManager notifies us that a 
// change has been made in the transformation controls (in other words - it tells
// us about a change in the shape of the bounding box).
// 
// The FreeTransformManager does NOT directly alter any SDT sprites.  It's up to
// us to decided whether or not to use the incoming information and (if so) how
// to apply it to the scene.
function onTransform(evt:FreeTransformEvent):void {
	// The user has made a change.  For the moment, the transformation library "knows"
	// the details of the transformation, but its knowledge is ephemeral.  We need to 
	// capture this information and store it in a persistent data structure.
	
	// Once the transformation details have been stored, they can be used during 
	// subsequent draw-cycles to ensure that the transformation is continuously
	// reflected in the stage contents (in other words - it becomess a persistent
	// part of the user's gameplay experience).
	
	// Store the transformation details
	if (evt.targetObject == leftBreast) {
		mostRecentTransform_LeftBreast = evt.matrix.clone();
	}
	else if (evt.targetObject == rightBreast) {
		mostRecentTransform_RightBreast = evt.matrix.clone();
	}
	
	// The matrix can now be carefully re-asserted during each frame-draw, AFTER the
	// SDT engine has done its work.  Yes, this is every bit as tedious as it sounds.  But it works.
	
	// Special rule: the breastSculpt mod applies is supposed to apply its transformation
	// effects to both of the breasts.  Whenever we detect a rightBreast change we'll
	// mirror it onto the left.
	if (evt.targetObject == rightBreast) {
		onTransform(new com.ryan.geom.FreeTransformEvent(evt.type, leftBreast, evt.matrix));
	}
	// HOOK: relax/rescind this rule if we ever decide to support "lopsided" sculpting.
}

// This function is called when the player presses the cancellation key.
// It aborts any in-progress sculpting operation and attempts to rollback
// to the saved "reference point".
function cancelBreastSculpting()
{
	// Perform the rollback
	anatomyTransform.cancelTransformation();
}

// This function is called when the player presses the standard hotkey.
// It initiates a sculpting operation in which only proportional changes
// (scaling, rotation, and translation) are allowed.
// TODO: this function is misnamed at present.  If the pro/dispro distinction
//       becomes relevant, we should reconsider the function names.
function toggleBreastSculpting_Proportionate()
{
	toggleBreastSculpting(rightBreast, true, true, true, true, true, false);
	// TODO: config settings.  Each category could be independently allowed or disallowed.
}

// This function is called when the player presses the secondary hotkey.
// It initiates a sculpting operation in which disproportionate changes 
// (skew and axial stretch) are permitted.
// Note: these changes can distort the image, but they still constitute
// an affine transformation.  Trapezoidal deformation is handled elsewhere.
//
// TODO: update this comment when trapezoidal deformation is implemented.
// TODO: this function is disabled as of V2
function toggleBreastSculpting_Disproportionate()
{
	//toggleBreastSculpting(rightBreast, false, false, true, false);
}

// This function is called when the user presses the secondary hotkey.
// It initiates a non-transformative UI mode, in which the only permitted
// operation is the movement of the center point.
// This is intended to allow fine-tuning of the axis of transformation,
// without the UI "clutter" which is present during full expert-mode
// operations.
function toggleBreastSculpting_Recenter()
{
	toggleBreastSculpting(rightBreast, false, false, false, false, false, true);
}

// This function toggles the sculpting mode.  If a sculpting operation is underway
// (e.g. bounding box and drag handles are visible) then we'll consider the operation
// complete (hide the sculpting controls; resume normal gameplay).
//
// If no sculpting operation is underway, then we'll draw a set of UI elements
// appropriate to the function parameters (e.g. rotation only?  okay, we'll show
// the "corner" drag handles while hiding the "edge" drag handles).
//
// CAUTION: We *expect* that this function will be invoked mainly by keystrokes,
// and that these will occur in symmetric pairs (e.g. a keystroke to show the menu,
// a few mouse operations, then a keystroke to close the menu).
// However, we must NOT assume that this is true!  It is possible to open the 
// transformation menu in other ways (such as by clicking on the breast itself,
// assuming that it has been registered for such interaction).
// This is a "toggle" function in name and in purpose, but we must not make hasty
// assumptions (such as attempting to infer game-state information based on the
// number of times that this function has been invoked).
// 
// LIMITATION: as of V1, there is no separation between the various transformation
// modes (rotation, rescale, skew) - they are all allowed, and all handles are visible.
// User feedback will be used to determine whether mode-limited sculpting is a necessary feature.
// In other words, the function parameters are currently superfluous.
function toggleBreastSculpting(targetObject:DisplayObject, allowResize:Boolean=true, allowRescale:Boolean=true, allowRotation:Boolean=true, allowSkew:Boolean=false, allowTranslate:Boolean=true, allowRecenter:Boolean=false)
{
	// The user has toggled the mod's activity state (presumably by pressing the hotkey).
	// First, we need to determine whether the activity state is being toggled ON or OFF.
	if(anatomyTransform.noActiveTarget && (allowResize || allowRescale || allowRotation || allowSkew || allowTranslate || allowRecenter))
	{
		// The transformation code does not have an active target (nothing is "in focus").
		// In other words, "sculpt mode" is currently inactive and we need to activate it.
		
		// Make the transformation bounding box (and drag handles) appear, surrounding the breasts of the 
		// HER model.  The transformation library will automatically choose a conformal shape.

		// Assume that we've transformed the target DisplayObject before.  Supply the
		// sculpt matrix to the transformation library so that it can serve as the 
		// "basis" for further drag-and-drop operations.
		var mostRecentSculptMatrix:Matrix = mostRecentTransform_RightBreast.clone();

		// Although the overall shapes are overlapping (that is, the bounding box completely encompasses
		// the breasts), they will not be concentric.  To ensure that rotation and skew operations are
		// performed with respect to the correct center point, we'll forcibly re-align the bounding box.
		anatomyTransform.activateSprite(targetObject, breastOffsetMatrix, mostRecentSculptMatrix, allowRotation, allowResize, allowRescale, allowSkew, allowTranslate, allowRecenter);
//main.updateStatusCol("blah","#00ff00");
		// At this point, we've drawn the bounding box (and handles) on-screen.  Everything
		// is in an appropriate  location (either "default position" or "outcome of most recent
		// transform" position).  We're ready to process the player's decisions.
		
		// Allow the transformation UI elements (drag handles) to receive mouse input
		anatomyTransform.mouseEnabled = true;
	}
	else
	{
		// The action is complete; release the "focus" and hide the transformation-related UI elements
		anatomyTransform.deactivateTarget();
		// Prevent the transformation object from receiving mouse input.  If we omit this step
		// then the transformation box will reappear as soon as the user clicks on a breast.
		anatomyTransform.mouseEnabled = (allowMouseActivation == 1);
	}
}


/*
 * The upcoming functions form the "heart" of this mod.  Each one will be invoked automatically 
 * by the SDT Loader's lProxy feature, in conjunction with invocations of the associated function
 * of the core SDT code.
 * Whether the mod function is invoked BEFORE or AFTER its corresponding core function is a 
 * design-time (hardcoded) decision, and is based on various factors (such as data integrity, 
 * inter-mod compatability, etc).
 *
 * It is very important to note that the SDT core achieves smooth animation by actively modifying
 * breast shape parameters.  As the HER model moves, minor changes are automatically made in the 
 * scale, position, skew, and rotation of both breasts.  These changes take the form of transform
 * parameter overwrites (as opposed to increment/decrement operations).
 *
 * Hence, we cannot simply perform a change (such as leftBreast.scaleX = 1.25) as a one-time
 * action in response to a UI operation.  Instead, we must continuously re-assert the desired 
 * transformation, because SDT core code is constantly trying to "wipe it away".
*/



// The following two functions are identical in purpose and effect; each applies to a different
// breast object.  These functions will be called immediately AFTER the breastController.Update
// function is invoked by core SDT logic (which will occur during most frames of gameplay,
// but it won't occur while the game is paused or while the user is making adjustments via
// the in-game menus).
//
// The update() function will wipe away any previous transformations that we've applied to the
// breast objects, assigning them a shape based on the core SDT logic (breathing cycle, linear
// squash animation, etc).  We will accept this as the "baseline" shape, and then apply our
// sculpting transformation on top of it.
//
// The result is that the final breast shape (as seen by the user) will be both sculpted and 
// animated.
function leftBreastUpdate_Post()
{
	(her.leftBreastController._breastMatrix as Matrix).concat(mostRecentTransform_LeftBreast);
	leftBreast.transform.matrix = (her.leftBreastController._breastMatrix as Matrix);
	// Note: I'd prefer to just call the concat function on the leftBreast.matrix.transform
	// "in-place."  Unfortunately, that doesn't work.  ActionScript ignores the change.
	// So we concat the underlying variable and then re-apply it onto the object.  Meh; it works.
	
	// Apply this same change to subsidiary clothing layers (e.g. bra cups)
	var _breastCostumeLayer:Object;
	for each (_breastCostumeLayer in her.leftBreastController._costumeLayers)
	{
		(_breastCostumeLayer as DisplayObject).transform.matrix = leftBreast.transform.matrix;
	}
}

function rightBreastUpdate_Post()
{
	(her.rightBreastController._breastMatrix as Matrix).concat(mostRecentTransform_RightBreast);
	rightBreast.transform.matrix = (her.rightBreastController._breastMatrix as Matrix);
	// Note: I'd prefer to just call the concat function on the leftBreast.matrix.transform
	// "in-place."  Unfortunately, that doesn't work.  ActionScript ignores the change.
	// So we concat the underlying variable and then re-apply it onto the object.  Meh; it works.
	
	// Apply this same change to subsidiary clothing layers (e.g. bra cups)
	var _breastCostumeLayer:Object;
	for each (_breastCostumeLayer in her.rightBreastController._costumeLayers)
	{
		(_breastCostumeLayer as DisplayObject).transform.matrix = rightBreast.transform.matrix;
	}
}

// This proxy function will fire AFTER a change is made in the breast size+shape (usually via the 
// in-game slider, but it also occurs while loading a new character).  Whenever such a change occurs,
// the game will redraw the inframammary fold (with a size/shape/thickness/colour which is approriate
// to the selected breast type).
//
// By reacting immediately AFTER the change, we can reassert the preferred visibility. If it's
// supposed to be hidden, then we'll re-hide it before the user even sees the unwanted sprites 
// for a single frame.
function breastSizeChanged_Post()
{
	// Reassert the preferred visibility setting for the inframammary fold
	helper_InframammaryVisibility();
}

function getSwfModsDataString_Pre():String
{
	try {
		return new String(";" + charData_LB + ":" + mostRecentTransform_LeftBreast.toString() + ";" + charData_RB + ":" + mostRecentTransform_RightBreast.toString() + ";" + charData_IM + ":" + hideInframammaryFold);
	} catch (myError:Error) {}
	return "";
}

// This is a proxy function which fires BEFORE the SDT core logic attempts to 
// load a character from a Character Data string (which is a long semicolon-
// delimited string without line breaks or whitespace).
// We will "intercept" this string and attempt to parse out any breastSculpt
// parameters which may have been "snuck in" when the Character Data string
// was saved.
//
// NOTE: for this purpose, it doesn't really matter whether this function
// fires BEFORE or AFTER the core loadData(...) function.  However, the
// lProxy class provides actual invocation parameters only to PRE functions.
// Therefore we must go BEFORE, because it's the only way that we'll have
// access to the Character Data string.
//
// NOTE: this function has a void return type.  This is important!  Since we're
// a PRE function, if we actually return a value then lProxy will refrain from
// invoking the original function (thus, the new character won't actually load
// and we've seriously disrupted someone's ability to play the game!)
function loadData_Pre(charData:String="", param2:Boolean=true, param3:Boolean=false, param4:Object=null):void
{
	// Parse the Character Data into a dictionary
	var parameters:Array = charData.split(";");
	var paramDictionary:Dictionary = new Dictionary();
	var parameter:String;
	for each (parameter in parameters)
	{
		try {
			paramDictionary[parameter.split(":")[0]] = parameter.split(":")[1];
		} catch (myError:Error) { /* Ignore any parsing failures */ }
	}

	// If we can find usable matrices in the CharData then go ahead and load them.
	// As a safeguard against incomplete or corrupted data, we'll load each set
	// of transformation parameters separately.
	try {
		mostRecentTransform_LeftBreast = helper_MatrixFromString(paramDictionary[charData_LB]);
	} catch (myError:Error) {
		// If the LOAD data lacks sculpting details, then we'll either retain the
		// pre-existing sculpting matrix, or revert to a clean-slate (depends on setting).
		if (retainCustomSculpting == 0) { mostRecentTransform_LeftBreast = new Matrix(); }
	}
	try {
		mostRecentTransform_RightBreast = helper_MatrixFromString(paramDictionary[charData_RB]);
	} catch (myError:Error) {
		// If the LOAD data lacks sculpting details, then we'll either retain the
		// pre-existing sculpting matrix, or revert to a clean-slate (depends on setting).
		if (retainCustomSculpting == 0) { mostRecentTransform_RightBreast = new Matrix(); }	
	}
	try {
		var inframammarySetting:Number = parseInt(paramDictionary[charData_IM]);
		hideInframammaryFold = isNaN(inframammarySetting) ? hideInframammaryFold : inframammarySetting;
		helper_InframammaryVisibility();
	} catch (myError:Error) {}
//main.updateStatusCol("LOADED:" + mostRecentTransform_LeftBreast.toString(),"#0000FF"); 
}

// This is a proxy function which fires BEFORE the SDT core logic attempts to 
// save a character to a Character Data string (which is a long semicolon-
// delimited string without line breaks or whitespace).
// Specifically, we're intervening in the subfunction which retrieves mod data.
//
// If we return a value, then the lProxy class will treat it as the return
// value for the original function call (thus: the original function won't
// even be invoked).  If we return nothing,  lProxy will call the original
// function and use its return value normally.
// 
// Therefore, we need to decide (WITHIN this function) whether it's appropriate
// to shove our data into the save-data stream.
function getSaveDataString_Pre(includeModData:Boolean=true) : String {
	if (includeModData || !includeSculptDataInCharacter) {
		// Either the original function call has requested mod data, or
		// our settings indicate that we shouldn't include breastSculpt
		// data in saves.

		// Therefore - we do not need to intervene.  Return nothing so
		// that the lProxy class can invoke the original function.
		return null;
	}
	
	// If we've reached this point, then SDT is about to generate a set of mod-less
	// Character Data AND our settings indicate that Character Data is SUPPOSED to
	// include the breastSculpt parameters.
	
	// We want to invoke the original function with a different parameter (so that 
	// the result _WILL_ include the sculpt parameters).  Unfortunately, we can't invoke 
	// it right now - we would merely end up recurding into THIS function!
	//
	// Instead, we'll temporarily remove the proxy.
	getSaveDataString.removePre(getSaveDataString_Pre);

	// We can now invoke the original function and obtain the desired information.
	var charDataWithSculptData:String = g.inGameMenu.getSaveDataString(true);
	
	// Now, we'll reassert the proxy so that subsequent "Save Character" attempts get
	// caught and scrutinized by the code of THIS function.
	getSaveDataString.addPre(getSaveDataString_Pre, true);

	// We can now return the result.
	return charDataWithSculptData;
	// Summary: SDT code called the function "getSaveData(false)".
	// through various shenanigans, we've forced that function call to return
	// the results of "getSaveData(true)".  SDT core logic does not realize
	// that any substitution has occurred; it will save the result to HDD.
}

// Helper function; it's intended to reverse the Matrix.ToString() function.
function helper_MatrixFromString(matrixString:String):Matrix
{
	// Sanitation: strip away leading and trailing braces, and any whitespace
	matrixString = matrixString.replace(new RegExp(/[() ]/g), "");
	// Sanitation: shift to uppercase to eliminate ambiguity
	matrixString = matrixString.toUpperCase();
	
	var matrix:Matrix = new Matrix();

	var matrixComponents:Array = matrixString.split(",");
	var matrixComponent:String;
	var componentValue:Number;
	for each (matrixComponent in matrixComponents)
	{
		componentValue = parseFloat(matrixComponent.split("=")[1]);
		switch (matrixComponent.split("=")[0]) {
			case "A":
				matrix.a = componentValue;
				break;
			case "B":
				matrix.b = componentValue;
				break;
			case "C":
				matrix.c = componentValue;
				break;
			case "D":
				matrix.d = componentValue;
				break;
			case "X":
				// fall-through: treat X the same as TX
			case "TX":
				matrix.tx = componentValue;
				break;
			case "Y":
				// fall-through: treat Y the same as TY
			case "TY":
				matrix.ty = componentValue;
				break;
		}
	}
	return matrix;
}

// Helper function.  Changes the visibility of inframammary sprites based on config settings.
// This code was moved to a helper function in accord with the "rule of three."
function helper_InframammaryVisibility():void
{
	her.torso.midLayer.chest.getChildAt(1).visible = her.torso.midLayer.chest.getChildAt(2).visible = (hideInframammaryFold != 1);
}

function doUnload()
{
	anatomyTransform.removeEventListener(FreeTransformEvent.ON_TRANSFORM, onTransform);
	anatomyTransform.removeEventListener(FreeTransformEvent.ON_CANCEL, onTransform);
	anatomyTransform.removeEventListener(FreeTransformEvent.ON_FOCUS, onFocus);
//	delete anatomyTransform;
	anatomyTransform = null;
	// TODO: remove anatomyTransform from scene?
}